@@ -2,8 +2,11 @@ package daemon
 
 import (
 	"bytes"
+	"encoding/json"
+	"flag"
 	"fmt"
 	"io/ioutil"
+	"math/rand"
 	"os"
 	"os/exec"
 	"path/filepath"
@@ -16,13 +19,15 @@ import (
 	"github.com/golang/glog"
 	drain "github.com/openshift/kubernetes-drain"
 	goerr "github.com/pkg/errors"
+	corev1 "k8s.io/api/core/v1"
 	"k8s.io/apimachinery/pkg/api/errors"
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
 	v1 "k8s.io/apimachinery/pkg/apis/meta/v1"
+	"k8s.io/apimachinery/pkg/labels"
+	"k8s.io/apimachinery/pkg/types"
+	"k8s.io/apimachinery/pkg/util/strategicpatch"
 	"k8s.io/apimachinery/pkg/util/wait"
-
-	// "k8s.io/client-go/informers"
-	corev1 "k8s.io/api/core/v1"
+	"k8s.io/client-go/informers"
 	"k8s.io/client-go/kubernetes"
 	"k8s.io/client-go/tools/cache"
 
@@ -64,7 +69,12 @@ type Daemon struct {
 	mu *sync.Mutex
 }
 
-const scriptsPath = "/bindata/scripts/enable-rdma.sh"
+const (
+	scriptsPath  = "/bindata/scripts/enable-rdma.sh"
+	annoKey      = "sriovnetwork.openshift.io/state"
+	annoIdle     = "Idle"
+	annoDraining = "Draining"
+)
 
 var namespace = os.Getenv("NAMESPACE")
 var pluginsPath = os.Getenv("PLUGINSPATH")
@@ -95,13 +105,14 @@ func (dn *Daemon) Run(stopCh <-chan struct{}, exitCh <-chan error) error {
 	if err := tryCreateUdevRule(); err != nil {
 		return err
 	}
-
+	var timeout int64 = 5
 	dn.mu = &sync.Mutex{}
 	informerFactory := sninformer.NewFilteredSharedInformerFactory(dn.client,
-		time.Second*30,
+		time.Second*15,
 		namespace,
 		func(lo *v1.ListOptions) {
 			lo.FieldSelector = "metadata.name=" + dn.name
+			lo.TimeoutSeconds = &timeout
 		},
 	)
 
@@ -111,8 +122,23 @@ func (dn *Daemon) Run(stopCh <-chan struct{}, exitCh <-chan error) error {
 		UpdateFunc: dn.nodeStateChangeHandler,
 	})
 
+	cfgInformerFactory := sninformer.NewFilteredSharedInformerFactory(dn.client,
+		time.Second*30,
+		namespace,
+		func(lo *v1.ListOptions) {
+			lo.FieldSelector = "metadata.name=" + "default"
+		},
+	)
+
+	cfgInformer := cfgInformerFactory.Sriovnetwork().V1().SriovOperatorConfigs().Informer()
+	cfgInformer.AddEventHandler(cache.ResourceEventHandlerFuncs{
+		AddFunc:    dn.operatorConfigAddHandler,
+		UpdateFunc: dn.operatorConfigChangeHandler,
+	})
+
 	time.Sleep(5 * time.Second)
 	go informer.Run(dn.stopCh)
+	go cfgInformer.Run(dn.stopCh)
 
 	for {
 		select {
@@ -130,6 +156,19 @@ func (dn *Daemon) Run(stopCh <-chan struct{}, exitCh <-chan error) error {
 	}
 }
 
+func (dn *Daemon) operatorConfigAddHandler(obj interface{}) {
+	dn.operatorConfigChangeHandler(&sriovnetworkv1.SriovOperatorConfig{}, obj)
+}
+
+func (dn *Daemon) operatorConfigChangeHandler(old, new interface{}) {
+	newCfg := new.(*sriovnetworkv1.SriovOperatorConfig)
+	var level = glog.Level(newCfg.Spec.LogLevel)
+	if level != flag.Lookup("v").Value.(flag.Getter).Get() {
+		glog.Infof("Set log verbose level to: %d", level)
+		flag.Set("v", level.String())
+	}
+}
+
 func (dn *Daemon) nodeStateAddHandler(obj interface{}) {
 	// "k8s.io/apimachinery/pkg/apis/meta/v1" provides an Object
 	// interface that allows us to get metadata easily
@@ -470,6 +509,45 @@ func (a GlogLogger) Logf(format string, v ...interface{}) {
 	glog.Infof(format, v...)
 }
 
+func (dn *Daemon) annotateNode(node, value string) error {
+	glog.Infof("annotateNode(): Annotate node %s with: %s", node, value)
+
+	err := wait.PollImmediate(10*time.Second, 1*time.Minute, func() (bool, error) {
+		oldNode, err := dn.kubeClient.CoreV1().Nodes().Get(dn.name, metav1.GetOptions{})
+		if err != nil {
+			glog.Infof("annotateNode(): Failed to get node %s %v, retrying", node, err)
+			return false, nil
+		}
+		oldData, err := json.Marshal(oldNode)
+		if err != nil {
+			return false, err
+		}
+
+		newNode := oldNode.DeepCopy()
+		if newNode.Annotations == nil {
+			newNode.Annotations = map[string]string{}
+		}
+		if newNode.Annotations[annoKey] != value {
+			newNode.Annotations[annoKey] = value
+			newData, err := json.Marshal(newNode)
+			if err != nil {
+				return false, err
+			}
+			patchBytes, err := strategicpatch.CreateTwoWayMergePatch(oldData, newData, corev1.Node{})
+			if err != nil {
+				return false, err
+			}
+			_, err = dn.kubeClient.CoreV1().Nodes().Patch(dn.name, types.StrategicMergePatchType, patchBytes)
+			if err != nil {
+				glog.Infof("annotateNode(): Failed to patch node %s %v, retrying", node, err)
+				return false, nil
+			}
+		}
+		return true, nil
+	})
+	return err
+}
+
 func (dn *Daemon) drainNode(name string) {
 	glog.Info("drainNode(): Update prepared")
 	var node *corev1.Node
@@ -488,6 +566,50 @@ func (dn *Daemon) drainNode(name string) {
 		return
 	}
 
+	rand.Seed(time.Now().UnixNano())
+	informerFactory := informers.NewSharedInformerFactory(dn.kubeClient,
+		time.Second*15,
+	)
+	stop := make(chan struct{})
+	lister := informerFactory.Core().V1().Nodes().Lister()
+	informer := informerFactory.Core().V1().Nodes().Informer()
+	informer.AddEventHandler(cache.ResourceEventHandlerFuncs{
+		UpdateFunc: func(oldObj, newObj interface{}) {
+			nn := newObj.(*corev1.Node)
+			if nn.Annotations[annoKey] == annoDraining || nn.GetName() == dn.name {
+				return
+			}
+			ok := true
+			// wait a random time to avoid all the nodes checking the nodes anno at the same time
+			time.Sleep(time.Duration(rand.Intn(6000)) * time.Millisecond)
+			nodes, err := lister.List(labels.Everything())
+			if err != nil {
+				return
+			}
+			glog.Infof("drainNode(): Check if any other node is draining")
+			for _, node := range nodes {
+				if node.GetName() != dn.name && node.Annotations[annoKey] == annoDraining {
+					glog.Infof("drainNode(): node %s is draining", node.Name)
+					ok = false
+				}
+			}
+			if ok {
+				glog.Info("drainNode(): No other node is draining, stop watching")
+				select {
+				case <-stop:
+				default:
+					close(stop)
+				}
+			}
+		},
+	})
+	informer.Run(stop)
+
+	err = dn.annotateNode(dn.name, annoDraining)
+	if err != nil {
+		glog.Errorf("drainNode(): Failed to annotate node: %v", err)
+	}
+
 	backoff := wait.Backoff{
 		Steps:    5,
 		Duration: 10 * time.Second,
@@ -497,11 +619,12 @@ func (dn *Daemon) drainNode(name string) {
 
 	logger := GlogLogger{}
 
+	glog.Info("drainNode(): Start draining")
 	if err := wait.ExponentialBackoff(backoff, func() (bool, error) {
 		err := drain.Drain(dn.kubeClient, []*corev1.Node{node}, &drain.DrainOptions{
 			DeleteLocalData:    true,
 			Force:              true,
-			GracePeriodSeconds: 600,
+			GracePeriodSeconds: -1,
 			IgnoreDaemonsets:   true,
 			Logger:             logger,
 		})
@@ -518,6 +641,10 @@ func (dn *Daemon) drainNode(name string) {
 		glog.Errorf("drainNode(): failed to drain node: %v", err)
 	}
 	glog.Info("drainNode(): drain complete")
+	err = dn.annotateNode(dn.name, annoIdle)
+	if err != nil {
+		glog.Errorf("drainNode(): failed to annotate node: %v", err)
+	}
 }
 
 func registerPlugins(ns *sriovnetworkv1.SriovNetworkNodeState) []string {
